/*
    Copyright (C) 2005-2012  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <glib-object.h>
#include "nntpgrab.h"
#include "collections.h"

#define IDLE_PRIORITY   G_PRIORITY_HIGH_IDLE + 50

NntpgrabCore *get_core(void);

#ifdef IN_NNTPGRAB_CORE
#include "plugins.h"

#define emit_event(obj, event_name, params...)              \
        g_signal_emit_by_name(obj, event_name, ## params);  \
        ng_plugin_emit_internal_event(event_name, ## params);
#else
#define emit_event(obj, event_name, params...)              \
        g_signal_emit_by_name(obj, event_name, ## params);
#endif

struct _do_emit_connecting_data {
    char servername[128];
    int conn_id;
};

struct _do_emit_connected_data {
    char servername[128];
    int conn_id;
    char welcome_msg[1024];
};

struct _do_emit_disconnect_data {
    char servername[128];
    int conn_id;
    NNTPDisconnectType disconnect_type;
    char errmsg[1024];
};

struct _traffic_data {
    int bytes_received[10];
    guint64 stamp;
};

struct _part_data {
    char servername[128];
    int conn_id;
    char collection_name[256];
    char subject[256];
    int part_num;
    int size;
    gboolean all_servers_tried;
};

struct _part_progress_data {
    char servername[128];
    int conn_id;
    char collection_name[256];
    char subject[256];
    int part_num;
    int bytes_downloaded;
    int bytes_total;
};

struct _collection_modified_data {
    char collection_name[256];
    char poster[256];
};

struct _file_added_data {
    char collection_name[256];
    char subject[256];
    char poster[256];
    guint64 stamp;
    guint64 file_size;
    guint64 total_file_size;
    guint64 total_file_size_remaining;
    int num_parts;
    NGTaskState status;
    GList *groups;
};

struct _file_removed_data {
    char collection_name[256];
    char subject[256];
    guint64 total_file_size;
    guint64 total_file_size_remaining;
};

struct _file_download_state_update_data {
    char collection_name[256];
    char subject[256];
    int num_parts_total;
    int num_parts_done;
    int num_parts_failed;
    guint64 file_size;
    guint64 file_size_remaining;
    guint64 total_file_size;
    guint64 total_file_size_remaining;
};

struct _file_state_data {
    char collection_name[256];
    char subject[256];
    char real_filename[256];
    NGTaskState old_state;
    NGTaskState new_state;
    guint64 file_size_remaining;
    guint64 total_size;
    guint64 total_size_remaining;
};

struct _schedular_data {
    NGSchedularState state;
    char reason[1024];
};

struct _log_data {
    char component[64];
    NGLogLevel log_level;
    char message[1024];
    gboolean emit_to_frontend;
    gboolean emit_to_plugins;
};

struct _move_data {
    char orig_collection_name[256];
    char new_collection_name[256];
    char subject[256];
    int old_position;
    int new_position;
};

struct _plugin_load_data {
    char plugin_name[32];
    gboolean is_persistent;
};

struct _plugin_event_data {
    char plugin_name[32];
    char event_name[32];
    char **params;
};

typedef struct _signal_data {
    union {
        struct _do_emit_connecting_data connecting_data;
        struct _do_emit_connected_data connected_data;
        struct _do_emit_disconnect_data disconnect_data;
        struct _traffic_data traffic_data;
        struct _part_data part_data;
        struct _part_progress_data part_progress_data;
        struct _collection_modified_data collection_modified_data;
        struct _file_added_data file_added_data;
        struct _file_removed_data file_removed_data;
        struct _file_download_state_update_data file_download_state_update_data;
        struct _file_state_data file_state_data;
        struct _schedular_data schedular_data;
        struct _log_data log_data;
        struct _move_data move_data;
        struct _plugin_load_data plugin_load_data;
        struct _plugin_event_data plugin_event_data;
    };
} SignalData;

static GTrashStack *signal_data_stack = NULL;
static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
static volatile int num_queued_part_progress_events = 0;

static SignalData *
pop_signal_data(void)
{
    SignalData *ret = NULL;

    g_static_mutex_lock(&mutex);

    if (signal_data_stack) {
        ret = g_trash_stack_pop(&signal_data_stack);
        memset(ret, 0, sizeof(SignalData));
    }

    if (!ret) {
        ret = g_slice_new0(SignalData);
    }

    g_static_mutex_unlock(&mutex);

    return ret;
}

static void
push_signal_data(SignalData *signal_data)
{
    g_static_mutex_lock(&mutex);

    g_trash_stack_push(&signal_data_stack, signal_data);

    g_static_mutex_unlock(&mutex);
}

void
core_signals_clear_stack(void)
{
    SignalData *data;

    g_static_mutex_lock(&mutex);

    while ((data = g_trash_stack_pop(&signal_data_stack))) {
        g_slice_free(SignalData, data);
    }

    g_static_mutex_unlock(&mutex);
}

static gboolean
do_emit_config_changed(gpointer data)
{
    if (!get_core()) {
        return FALSE;
    }

    emit_event(get_core(), "config_changed");

    return FALSE;
}

void
nntpgrab_core_emit_config_changed(gboolean direct)
{
    if (!get_core()) {
        return;
    }

    if (direct) {
        do_emit_config_changed(NULL);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_config_changed, NULL, NULL);
    }
}

static gboolean
do_emit_all_downloads_completed(gpointer data)
{
    if (!get_core()) {
        return FALSE;
    }

    emit_event(get_core(), "all_downloads_completed");

    return FALSE;
}

void
nntpgrab_core_emit_all_downloads_completed(gboolean direct)
{
    if (!get_core()) {
        return;
    }

    if (direct) {
        do_emit_all_downloads_completed(NULL);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_all_downloads_completed, NULL, NULL);
    }
}

static gboolean
do_emit_connecting(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "connection_connecting", signal_data->connecting_data.servername, signal_data->connecting_data.conn_id);
    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_connecting(gboolean direct, const char *servername, int conn_id)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->connecting_data.servername, servername, sizeof(data->connecting_data.servername) - 1);
    data->connecting_data.conn_id = conn_id;

    if (direct) {
        do_emit_connecting(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_connecting, data, NULL);
    }
}

static gboolean
do_emit_connected(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "connection_connected", signal_data->connected_data.servername, signal_data->connected_data.conn_id, signal_data->connected_data.welcome_msg);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_connected(gboolean direct, const char *servername, int conn_id, const char *welcome_msg)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->connected_data.servername, servername, sizeof(data->connected_data.servername) - 1);
    data->connected_data.conn_id = conn_id;
    strncpy(data->connected_data.welcome_msg, welcome_msg, sizeof(data->connected_data.welcome_msg) - 1);

    if (direct) {
        do_emit_connected(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_connected, data, NULL);
    }
}

static gboolean
do_emit_disconnect(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "connection_disconnect", signal_data->disconnect_data.servername, signal_data->disconnect_data.conn_id, signal_data->disconnect_data.disconnect_type, strlen(signal_data->disconnect_data.errmsg) > 0 ? signal_data->disconnect_data.errmsg : NULL);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_disconnect(gboolean direct, const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *errmsg)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->disconnect_data.servername, servername, sizeof(data->disconnect_data.servername) - 1);
    data->disconnect_data.conn_id = conn_id;
    data->disconnect_data.disconnect_type = disconnect_type;
    if (errmsg) {
        strncpy(data->disconnect_data.errmsg, errmsg, sizeof(data->disconnect_data.errmsg) - 1);
    }

    if (direct) {
        do_emit_disconnect(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_disconnect, data, NULL);
    }
}

static gboolean
do_emit_traffic_monitor_update(gpointer data)
{
    SignalData *signal_data = (SignalData*) data;
    double average;
    int i;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    // Take the average from the last 5 seconds
    average = 0;
    for (i = 5; i < 10; i++) {
        average += signal_data->traffic_data.bytes_received[i];
    }
    average /= 5;

    emit_event (get_core(), "traffic_monitor_update", signal_data->traffic_data.bytes_received[0],
                                                      signal_data->traffic_data.bytes_received[1],
                                                      signal_data->traffic_data.bytes_received[2],
                                                      signal_data->traffic_data.bytes_received[3],
                                                      signal_data->traffic_data.bytes_received[4],
                                                      signal_data->traffic_data.bytes_received[5],
                                                      signal_data->traffic_data.bytes_received[6],
                                                      signal_data->traffic_data.bytes_received[7],
                                                      signal_data->traffic_data.bytes_received[8],
                                                      signal_data->traffic_data.bytes_received[9],
                                                      signal_data->traffic_data.stamp,
                                                      average);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_traffic_monitor_update(gboolean direct, int bytes_received[10], time_t stamp)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    memcpy(data->traffic_data.bytes_received, bytes_received, sizeof(int) * 10);
    data->traffic_data.stamp = stamp;

    if (direct) {
        do_emit_traffic_monitor_update(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_traffic_monitor_update, data, NULL);
    }
}

static gboolean
do_emit_part_download_start(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event( get_core(), "part_download_start",  signal_data->part_data.servername,
                                                    signal_data->part_data.conn_id,
                                                    signal_data->part_data.collection_name,
                                                    signal_data->part_data.subject,
                                                    signal_data->part_data.part_num);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_download_start(gboolean direct, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->part_data.servername, servername, sizeof(data->part_data.servername) - 1);
    data->part_data.conn_id = conn_id;
    strncpy(data->part_data.collection_name, collection_name, sizeof(data->part_data.collection_name) - 1);
    strncpy(data->part_data.subject, subject, sizeof(data->part_data.subject) - 1);
    data->part_data.part_num = part_num;

    if (direct) {
        do_emit_part_download_start(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_part_download_start, data, NULL);
    }
}

static gboolean
do_emit_part_failed(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "part_failed",   signal_data->part_data.servername,
                                            signal_data->part_data.conn_id,
                                            signal_data->part_data.collection_name,
                                            signal_data->part_data.subject,
                                            signal_data->part_data.part_num,
                                            signal_data->part_data.size,
                                            signal_data->part_data.all_servers_tried);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_failed(gboolean direct, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gboolean all_servers_tried)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->part_data.servername, servername, sizeof(data->part_data.servername) - 1);
    data->part_data.conn_id = conn_id;
    strncpy(data->part_data.collection_name, collection_name, sizeof(data->part_data.collection_name) - 1);
    strncpy(data->part_data.subject, subject, sizeof(data->part_data.subject) - 1);
    data->part_data.part_num = part_num;
    data->part_data.all_servers_tried = all_servers_tried;
    data->part_data.size = size;

    if (direct) {
        do_emit_part_failed(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_part_failed, data, NULL);
    }
}

static gboolean
do_emit_part_done(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "part_done", signal_data->part_data.servername,
                                        signal_data->part_data.conn_id,
                                        signal_data->part_data.collection_name,
                                        signal_data->part_data.subject,
                                        signal_data->part_data.part_num,
                                        signal_data->part_data.size);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_done(gboolean direct, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->part_data.servername, servername, sizeof(data->part_data.servername) - 1);
    data->part_data.conn_id = conn_id;
    strncpy(data->part_data.collection_name, collection_name, sizeof(data->part_data.collection_name) - 1);
    strncpy(data->part_data.subject, subject, sizeof(data->part_data.subject) - 1);
    data->part_data.part_num = part_num;
    data->part_data.size = size;

    if (direct) {
        do_emit_part_done(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_part_done, data, NULL);
    }
}

static gboolean
do_emit_part_progress_update(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "part_progress_update",  signal_data->part_progress_data.servername,
                                                    signal_data->part_progress_data.conn_id,
                                                    signal_data->part_progress_data.collection_name,
                                                    signal_data->part_progress_data.subject,
                                                    signal_data->part_progress_data.part_num,
                                                    signal_data->part_progress_data.bytes_downloaded,
                                                    signal_data->part_progress_data.bytes_total);

    push_signal_data(signal_data);

    g_atomic_int_add(&num_queued_part_progress_events, -1);

    return FALSE;
}

void
nntpgrab_core_emit_part_progress_update(gboolean direct, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int bytes_downloaded, int bytes_total)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    /* Prevent too many queued events */
    if (num_queued_part_progress_events >= 20) {
        static time_t stamp_last_been_here = 0;

        if (stamp_last_been_here != time(NULL)) {
            g_print(__FILE__ ":%i %i Skipping part_progress_update event because too many events are already queued\n", __LINE__, (int) stamp_last_been_here);
            stamp_last_been_here = time(NULL);
        }

        return;
    }

    data = pop_signal_data();
    strncpy(data->part_progress_data.servername, servername, sizeof(data->part_progress_data.servername) - 1);
    data->part_progress_data.conn_id = conn_id;
    strncpy(data->part_progress_data.collection_name, collection_name, sizeof(data->part_progress_data.collection_name) - 1);
    strncpy(data->part_progress_data.subject, subject, sizeof(data->part_progress_data.subject) - 1);
    data->part_progress_data.part_num = part_num;
    data->part_progress_data.bytes_downloaded = bytes_downloaded;
    data->part_progress_data.bytes_total = bytes_total;

    g_atomic_int_inc(&num_queued_part_progress_events);

    if (direct) {
        do_emit_part_progress_update(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_part_progress_update, data, NULL);
    }
}

static gboolean
do_emit_collection_added(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "collection_added", signal_data->collection_modified_data.collection_name,
                                               signal_data->collection_modified_data.poster);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_collection_added(gboolean direct, const char *collection_name, const char *poster)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->collection_modified_data.collection_name, collection_name, sizeof(data->collection_modified_data.collection_name) - 1);
    strncpy(data->collection_modified_data.poster, poster, sizeof(data->collection_modified_data.poster) - 1);

    if (direct) {
        do_emit_collection_added(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_collection_added, data, NULL);
    }
}

static gboolean
do_emit_collection_removed(gpointer data)
{
    char *collection_name = (char *) data;

    if (!get_core()) {
        g_free(collection_name);
        return FALSE;
    }

    emit_event(get_core(), "collection_removed", collection_name);

    g_free(collection_name);

    return FALSE;
}

void
nntpgrab_core_emit_collection_removed(gboolean direct, const char *collection_name)
{
    if (!get_core()) {
        return;
    }

    if (direct) {
        do_emit_collection_removed(g_strdup(collection_name));
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_collection_removed, g_strdup(collection_name), NULL);
    }
}

static gboolean
do_emit_collection_modified(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "collection_modified", signal_data->collection_modified_data.collection_name,
                                                  signal_data->collection_modified_data.poster);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_collection_modified(gboolean direct, const char *collection_name, const char *poster)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->collection_modified_data.collection_name, collection_name, sizeof(data->collection_modified_data.collection_name) - 1);
    strncpy(data->collection_modified_data.poster, poster, sizeof(data->collection_modified_data.poster) - 1);

    if (direct) {
        do_emit_collection_modified(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_collection_modified, data, NULL);
    }
}

static gboolean
do_emit_file_added(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;
    GList *list;

    if (!get_core()) {
        list = signal_data->file_added_data.groups;
        while (list) {
            g_free(list->data);
            list = g_list_next(list);
        }

        g_list_free(signal_data->file_added_data.groups);
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "file_added", signal_data->file_added_data.collection_name,
                                         signal_data->file_added_data.subject,
                                         signal_data->file_added_data.poster,
                                         signal_data->file_added_data.stamp,
                                         signal_data->file_added_data.file_size,
                                         signal_data->file_added_data.total_file_size,
                                         signal_data->file_added_data.total_file_size_remaining,
                                         signal_data->file_added_data.status,
                                         signal_data->file_added_data.num_parts,
                                         signal_data->file_added_data.groups);

    list = signal_data->file_added_data.groups;
    while (list) {
        g_free(list->data);
        list = g_list_next(list);
    }

    g_list_free(signal_data->file_added_data.groups);
    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_added(gboolean direct, const char *collection_name, const char *subject, const char *poster, guint64 stamp, guint64 file_size, guint64 total_size, guint64 total_size_remaining, NGTaskState status, int num_parts, GList *groups)
{
    SignalData *data;
    GList *list;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->file_added_data.collection_name, collection_name, sizeof(data->file_added_data.collection_name) - 1);
    strncpy(data->file_added_data.subject, subject, sizeof(data->file_added_data.subject) - 1);
    strncpy(data->file_added_data.poster, poster, sizeof(data->file_added_data.poster) - 1);
    data->file_added_data.stamp = stamp;
    data->file_added_data.file_size = file_size;
    data->file_added_data.status = status;
    data->file_added_data.num_parts = num_parts;
    data->file_added_data.groups = NULL;
    data->file_added_data.total_file_size = total_size;
    data->file_added_data.total_file_size_remaining = total_size_remaining;

    list = groups;
    while (list) {
        char *group = (char *) list->data;
        data->file_added_data.groups = g_list_append(data->file_added_data.groups, g_strdup(group));
        list = g_list_next(list);
    }

    if (direct) {
        do_emit_file_added(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_file_added, data, NULL);
    }
}

static gboolean
do_emit_file_removed(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "file_removed",  signal_data->file_removed_data.collection_name,
                                            signal_data->file_removed_data.subject,
                                            signal_data->file_removed_data.total_file_size,
                                            signal_data->file_removed_data.total_file_size_remaining);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_removed(gboolean direct, const char *collection_name, const char *subject, guint64 total_file_size, guint64 total_file_size_remaining)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->file_removed_data.collection_name, collection_name, sizeof(data->file_removed_data.collection_name) - 1);
    strncpy(data->file_removed_data.subject, subject, sizeof(data->file_removed_data.subject) - 1);
    data->file_removed_data.total_file_size = total_file_size;
    data->file_removed_data.total_file_size_remaining = total_file_size_remaining;

    if (direct) {
        do_emit_file_removed(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_file_removed, data, NULL);
    }
}

static gboolean
do_emit_file_download_state_update(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "file_download_state_update", signal_data->file_download_state_update_data.collection_name,
                                                         signal_data->file_download_state_update_data.subject,
                                                         signal_data->file_download_state_update_data.num_parts_total,
                                                         signal_data->file_download_state_update_data.num_parts_done,
                                                         signal_data->file_download_state_update_data.num_parts_failed,
                                                         signal_data->file_download_state_update_data.file_size,
                                                         signal_data->file_download_state_update_data.file_size_remaining,
                                                         signal_data->file_download_state_update_data.total_file_size,
                                                         signal_data->file_download_state_update_data.total_file_size_remaining);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_download_state_update(gboolean direct, const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_file_size, guint64 total_file_size_remaining)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->file_download_state_update_data.collection_name, collection_name, sizeof(data->file_download_state_update_data.collection_name) - 1);
    strncpy(data->file_download_state_update_data.subject, subject, sizeof(data->file_download_state_update_data.subject) - 1);
    data->file_download_state_update_data.num_parts_total = num_parts_total;
    data->file_download_state_update_data.num_parts_done = num_parts_done;
    data->file_download_state_update_data.num_parts_failed = num_parts_failed;
    data->file_download_state_update_data.file_size = file_size;
    data->file_download_state_update_data.file_size_remaining = file_size_remaining;
    data->file_download_state_update_data.total_file_size = total_file_size;
    data->file_download_state_update_data.total_file_size_remaining = total_file_size_remaining;

    if (direct) {
        do_emit_file_download_state_update(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_file_download_state_update, data, NULL);
    }
}

static gboolean
do_emit_file_state_changed(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;
    char *real_filename;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    if (strlen(signal_data->file_state_data.real_filename) > 0) {
        real_filename = signal_data->file_state_data.real_filename;
    } else {
        real_filename = NULL;
    }

    emit_event(get_core(), "file_state_changed", signal_data->file_state_data.collection_name,
                                                 signal_data->file_state_data.subject,
                                                 real_filename,
                                                 signal_data->file_state_data.old_state,
                                                 signal_data->file_state_data.new_state,
                                                 signal_data->file_state_data.file_size_remaining,
                                                 signal_data->file_state_data.total_size,
                                                 signal_data->file_state_data.total_size_remaining);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_state_changed(gboolean direct, const char *collection_name, const char *subject, const char *real_filename, NGTaskState old_state, NGTaskState new_state, guint64 file_size_remaining, guint64 total_size, guint64 total_size_remaining)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->file_state_data.collection_name, collection_name, sizeof(data->file_state_data.collection_name) - 1);
    strncpy(data->file_state_data.subject, subject, sizeof(data->file_state_data.subject) - 1);
    if (real_filename) {
        strncpy(data->file_state_data.real_filename, real_filename, sizeof(data->file_state_data.real_filename) - 1);
    }
    data->file_state_data.old_state = old_state;
    data->file_state_data.new_state = new_state;
    data->file_state_data.file_size_remaining = file_size_remaining;
    data->file_state_data.total_size = total_size;
    data->file_state_data.total_size_remaining = total_size_remaining;

    if (direct) {
        do_emit_file_state_changed(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_file_state_changed, data, NULL);
    }
}

static gboolean
do_emit_schedular_state_changed(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "schedular_state_changed",   signal_data->schedular_data.state,
                                                        (strlen(signal_data->schedular_data.reason) > 0) ? signal_data->schedular_data.reason : NULL);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_schedular_state_changed(gboolean direct, NGSchedularState state, const char *reason)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    data->schedular_data.state = state;
    if (reason) {
        strncpy(data->schedular_data.reason, reason, sizeof(data->schedular_data.reason) - 1);
    }

    if (direct) {
        do_emit_schedular_state_changed(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_schedular_state_changed, data, NULL);
    }
}

static gboolean
do_emit_log_message(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    if (signal_data->log_data.emit_to_frontend) {
        g_signal_emit_by_name(get_core(),
                              "log_message",
                              signal_data->log_data.component,
                              signal_data->log_data.log_level,
                              signal_data->log_data.message);
    }

#ifdef IN_NNTPGRAB_CORE
    if (signal_data->log_data.emit_to_plugins) {
        ng_plugin_emit_internal_event("log_message",
                                      signal_data->log_data.component,
                                      signal_data->log_data.log_level,
                                      signal_data->log_data.message);
    }
#endif

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_log_message(gboolean direct, const char *component, NGLogLevel log_level, const char *message, gboolean emit_to_frontend, gboolean emit_to_plugins)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    g_return_if_fail(component != NULL);
    g_return_if_fail(message != NULL);

    data = pop_signal_data();
    strncpy(data->log_data.component, component, sizeof(data->log_data.component) - 1);
    data->log_data.log_level = log_level;
    strncpy(data->log_data.message, message, sizeof(data->log_data.message) - 1);
    data->log_data.emit_to_frontend = emit_to_frontend;
    data->log_data.emit_to_plugins = emit_to_plugins;

    if (direct) {
        do_emit_log_message(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_log_message, data, NULL);
    }
}

static gboolean
do_emit_file_moved(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "task_moved",    signal_data->move_data.orig_collection_name,
                                            signal_data->move_data.subject,
                                            signal_data->move_data.new_collection_name,
                                            signal_data->move_data.old_position,
                                            signal_data->move_data.new_position);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_moved(gboolean direct, const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->move_data.orig_collection_name, orig_collection_name, sizeof(data->move_data.orig_collection_name) - 1);
    strncpy(data->move_data.subject, subject, sizeof(data->move_data.subject) - 1);
    strncpy(data->move_data.new_collection_name, new_collection_name, sizeof(data->move_data.new_collection_name) - 1);
    data->move_data.old_position = old_position;
    data->move_data.new_position = new_position;

    if (direct) {
        do_emit_file_moved(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_file_moved, data, NULL);
    }
}

static gboolean
do_emit_collection_moved(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "collection_moved",  signal_data->move_data.orig_collection_name,
                                                signal_data->move_data.old_position,
                                                signal_data->move_data.new_position);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_collection_moved(gboolean direct, const char *collection_name, int old_position, int new_position)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->move_data.orig_collection_name, collection_name, sizeof(data->move_data.orig_collection_name) - 1);
    data->move_data.old_position = old_position;
    data->move_data.new_position = new_position;

    if (direct) {
        do_emit_collection_moved(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_collection_moved, data, NULL);
    }
}

static gboolean
do_emit_collection_downloaded(gpointer data)
{
    if (!get_core()) {
        g_free(data);
        return FALSE;
    }

    emit_event(get_core(), "collection_downloaded", (const char*) data);
    g_free(data);

    return FALSE;
}

void
nntpgrab_core_emit_collection_downloaded(gboolean direct, const char *collection_name)
{
    if (!get_core()) {
        return;
    }

    if (direct) {
        do_emit_collection_downloaded(g_strdup(collection_name));
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_collection_downloaded, g_strdup(collection_name), NULL);
    }
}

static gboolean
do_emit_plugin_loaded(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "plugin_loaded", signal_data->plugin_load_data.plugin_name,
                                            signal_data->plugin_load_data.is_persistent);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_plugin_loaded(gboolean direct, const char *plugin_name, gboolean is_persistent)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->plugin_load_data.plugin_name, plugin_name, sizeof(data->plugin_load_data.plugin_name) - 1);
    data->plugin_load_data.is_persistent = is_persistent;

    if (direct) {
        do_emit_plugin_loaded(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_plugin_loaded, data, NULL);
    }
}

static gboolean
do_emit_plugin_unloaded(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "plugin_unloaded", signal_data->plugin_load_data.plugin_name);

    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_plugin_unloaded(gboolean direct, const char *plugin_name)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->plugin_load_data.plugin_name, plugin_name, sizeof(data->plugin_load_data.plugin_name) - 1);

    if (direct) {
        do_emit_plugin_unloaded(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_plugin_unloaded, data, NULL);
    }
}

static gboolean
do_emit_plugin_event(gpointer data)
{
    SignalData *signal_data = (SignalData *) data;

    if (!get_core()) {
        g_strfreev(signal_data->plugin_event_data.params);
        push_signal_data(signal_data);
        return FALSE;
    }

    emit_event(get_core(), "plugin_event", signal_data->plugin_event_data.plugin_name,
                                           signal_data->plugin_event_data.event_name,
                                           signal_data->plugin_event_data.params);

    g_strfreev(signal_data->plugin_event_data.params);
    push_signal_data(signal_data);

    return FALSE;
}

void
nntpgrab_core_emit_plugin_event(gboolean direct, const char *plugin_name, const char *event_name, const char **params)
{
    SignalData *data;

    if (!get_core()) {
        return;
    }

    data = pop_signal_data();
    strncpy(data->plugin_event_data.plugin_name, plugin_name, sizeof(data->plugin_event_data.plugin_name) - 1);
    strncpy(data->plugin_event_data.event_name, event_name, sizeof(data->plugin_event_data.event_name) - 1);
    data->plugin_event_data.params = g_strdupv((gchar**)params);

    if (direct) {
        do_emit_plugin_event(data);
    } else {
        g_idle_add_full(IDLE_PRIORITY, do_emit_plugin_event, data, NULL);
    }
}
